#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Forecast data timeseries.

This module is an interface to the Salient `forecast_timeseries` API, which returns
probabilistic weather forecasts in subseasonal-to-seasonal timescales.

Command line usage example:

```
cd ~/salientsdk
# this will get a single variable in a single file:
python -m salientsdk forecast_timeseries -lat 42 -lon -73 --timescale seasonal -u username -p password
# this will get multiple variables in separate files:
python -m salientsdk forecast_timeseries -lat 42 -lon -73 -var temp,precip --timescale seasonal
```

"""

import os
from datetime import datetime

import requests
import xarray as xr

from . import login_api
from .constants import _build_url, get_model_version
from .location import Location
from .login_api import get_api_key, get_current_session, get_verify_ssl


def forecast_timeseries(
    loc: Location,
    date: str = "-today",
    debias: bool = False,
    field: str = "anom",
    format: str = "nc",
    model: str = "blend",
    reference_clim: str = "30_yr",
    timescale="all",
    variable: str = "temp",
    version: str = get_model_version(),
    force: bool = False,
    session: requests.Session = get_current_session(),
    apikey: str | None = get_api_key(),
    verify: bool = get_verify_ssl(),
    verbose: bool = False,
    **kwargs,
) -> str | dict[str, str]:
    """Get time series of S2S meteorological forecasts.

    This function is a convenience wrapper to the Salient
    [API](https://api.salientpredictions.com/v2/documentation/api/#/Forecasts/forecast_timeseries).

    Args:
        loc (Location): The location to query
        date (str): The date the forecast was generated.  Defaults to `-today`, which will find the
            most recent forecast.  Can also be a specific date in the format `YYYY-MM-DD`.
        debias (bool): If True, debias the data to local observations.
            Disabled for `shapefile` locations.
            [detail](https://salientpredictions.notion.site/Debiasing-2888d5759eef4fe89a5ba3e40cd72c8f)
        field (str): The field to query, defaults to `anom` which is an anomaly value from climatology.
            Also available: `vals`, which will return absolute values without regard to climatology.
        format (str): The file format of the response.
            Defaults to `nc` which returns a multivariate NetCDF file.
            Also available: `csv` which returns a CSV file.
        model (str): The model to query.  Defaults to `blend`, which is the Salient blended forecast.
        reference_clim (str):  Reference climatology for calculating anomalies.
            Ignored when `field=vals` since there are no anomalies to calculate.
            Defaults to `30_yr`, which is the 30-year climatology.
        timescale (str): Forecast look-ahead.
            - `sub-seasonal` is 1-5 weeks.  Will return a coordinate `forecast_date_weekly` and
                a data variable `anom_weekly` or `vals_weekly`.
            - `seasonal` is 1-3 months.  Will return a coordinate `forecast_date_monthly` and a
                data variable `anom_monthly` or `vals_monthly`.
            - `long-range` is 1-4 quarters.  Will return a coordinate `forecast_date_quarterly` and a
                data variable `anom_quarterly` or `vals_quarterly`.
            - `all` (default) will include `sub-seasonal`, `seasonal`, and `long-range` timescales
        variable (str): The variable to query, defaults to `temp`
            To request multiple variables, separate them with a comma `temp,precip`
            This will download one file per variable
            See the
            [Data Fields](https://salientpredictions.notion.site/Variables-d88463032846402e80c9c0972412fe60)
            documentation for a full list of available historical variables.
        version (str): The model version of the Salient `blend` forecast.
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided
            and `get_api_key()` returns `None`.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        **kwargs: Additional arguments to pass to the API

    Keyword Arguments:
        units (str): `SI` or `US`

    Returns:
        str | dict: the file name of the downloaded data.
            File names are a hash of the query parameters.
            When `force=False` and the file already exists, the function will return the file name
            almost instantaneously without querying the API.
            If multiple variables are requested, returns a `dict` of `{variable:file_name}`
    """
    assert field in [
        "anom",
        "vals",
        "vals_ens",
    ], f"Invalid field {field}"
    assert format in ["nc", "csv"], f"Invalid format {format}"

    if date == "-today":
        date = datetime.today().strftime("%Y-%m-%d")

    # if there is a comma in variable, vectorize:
    if isinstance(variable, str) and "," in variable:
        variable = variable.split(",")

    if isinstance(variable, list):
        file_names = {
            var: forecast_timeseries(
                loc=loc,
                date=date,
                debias=debias,
                field=field,
                format=format,
                model=model,
                reference_clim=reference_clim,
                timescale=timescale,
                variable=var,
                version=version,
                force=force,
                session=session,
                verify=verify,
                verbose=verbose,
                apikey=apikey,
                **kwargs,
            )
            for var in variable
        }
        if verbose:
            print(file_names)
        return file_names

    endpoint = "forecast_timeseries"
    args = loc.asdict(
        date=date,
        debias=debias,
        field=field,
        format=format,
        model=model,
        reference_clim=reference_clim,
        timescale=timescale,
        variable=variable,
        version=version,
        apikey=apikey,
        **kwargs,
    )

    (query, file_name) = _build_url(endpoint, args)

    if force or not os.path.exists(file_name):
        if verbose:
            print(f"Downloading {query} to {file_name}")
        with open(file_name, "wb" if format == "nc" else "w") as f:
            result = session.get(query, verify=verify)
            result.raise_for_status()
            if format == "nc":
                f.write(result.content)
            else:
                f.write(result.text)
    elif verbose:
        print(f"File {file_name} already exists")

    return file_name



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Constants for the Salient SDK.

This module contains constants used throughout the Salient SDK.

"""

import datetime
import hashlib
import itertools
import urllib

import pandas as pd
import requests

# This is the base URL for the Salient API:
URL = "https://api.salientpredictions.com/"

API_VERSION = "v2"

MODEL_VERSION = "v8"
MODEL_VERSIONS = ["v7_1", "v8"]


def _build_url(endpoint: str, args: None | dict = None) -> tuple[str, str]:
    url = URL + API_VERSION + "/" + endpoint
    file_name = endpoint

    if args:
        # apikey will often be None when we're using a persistent session:
        args = {k: v for k, v in args.items() if v is not None}

        url += "?"
        url += urllib.parse.urlencode(args, safe=",")

        # apikey doesn't influence the file contents, so shouldn't be in the hash:
        if "apikey" in args:
            del args["apikey"]

        file_name += "_"
        file_name += hashlib.md5(str(args).encode()).hexdigest()

        if "format" in args:
            file_name += "." + args["format"]

    return (url, file_name)


def _build_urls(endpoint: str, args: None | dict = None) -> pd.DataFrame:
    url = URL + API_VERSION + "/" + endpoint
    file_name = endpoint

    if args:
        # apikey will often be None when we're using a persistent session:
        args = {k: v for k, v in args.items() if v is not None}

        scalar_args = {k: v for k, v in args.items() if not isinstance(v, (list, tuple))}
        vector_args = {k: v for k, v in args.items() if isinstance(v, (list, tuple))}

        if vector_args:
            expanded_args = list(itertools.product(*vector_args.values()))
            expanded_args = [dict(zip(vector_args.keys(), values)) for values in expanded_args]
            queries = [
                _build_urls(endpoint, {**arg, **scalar_args}).assign(**arg)
                for arg in expanded_args
            ]
            return pd.concat(queries, ignore_index=True)

        url += "?"
        url += urllib.parse.urlencode(scalar_args, safe=",")

        # apikey doesn't influence the file contents, so shouldn't be in the hash:
        if "apikey" in args:
            del args["apikey"]

        file_name += "_"
        file_name += hashlib.md5(str(args).encode()).hexdigest()

        if "format" in args:
            file_name += "." + args["format"]

    return pd.DataFrame([{"query": url, "file_name": file_name}])


def _validate_date(date: str | datetime.datetime) -> str:
    if isinstance(date, str) and date == "-today":
        date = datetime.datetime.today()

    if isinstance(date, datetime.datetime):
        date = date.strftime("%Y-%m-%d")

    # ENHANCEMENT: accept other date formats like numpy datetime64, pandas Timestamp, etc
    # ENHANCEMENT: make sure date is properly formatted

    return date


def get_model_version() -> str:
    """Get the current default model version.

    Returns:
        str: The current model version

    """
    return MODEL_VERSION


def set_model_version(version: str) -> None:
    """Set the default model version.

    Args:
        version (str): The model version to set

    """
    assert version in MODEL_VERSIONS
    global MODEL_VERSION
    MODEL_VERSION = version



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Define the Location class."""

# Usage example:
# python location.py --lat 42 --lon -73

import argparse
import datetime
import os
import urllib

from . import constants, login_api


class Location:
    """Define geographical parameters for an API query.

    The Salient API
    can define location by a single latitude/longitude, multiple
    latitude/longitude pairs in a single location_file, or a polygon
    defined in a shapefile.
    """

    def __init__(
        self,
        lat: float | None = None,
        lon: float | None = None,
        location_file: str | None = None,
        shapefile: str | None = None,
        region: str | None = None,
    ):
        """Initialize a Location object.

        Only one of the following 4 options should be used at a time: lat/lon,
        location_file, shapefile, or region.

        Args:
            lat (float): Latitude in degrees, -90 to 90.
            lon (float): Longitude in degrees, -180 to 180.
            location_file (str): Path to a CSV file with latitude and longitude columns.
            shapefile (str): Path to a shapefile with a polygon defining the location.
            region (str): Accepts continents, countries, or U.S. states (e.g. "usa")
                Only available for `hindcast_summary()`
        """
        self.lat = lat
        self.lon = lon
        self.location_file = location_file
        self.shapefile = shapefile
        self.region = region

        self.__validate()

    def asdict(self, **kwargs) -> dict:
        """Render as a dictionary.

        Generates a dictionary representation that can be encoded into a URL.
        Will contain one and only one location_file, shapefile, or lat/lon pair.

        Args:
            **kwargs: Additional key-value pairs to include in the dictionary.
                Will validate some common arguments that are shared across API calls.
        """
        if self.location_file:
            dct = {"location_file": self.location_file, **kwargs}
        elif self.shapefile:
            dct = {"shapefile": self.shapefile, **kwargs}
        elif self.region:
            dct = {"region": self.region, **kwargs}
        else:
            dct = {"lat": self.lat, "lon": self.lon, **kwargs}

        if "apikey" in dct and dct["apikey"] is not None:
            assert isinstance(dct["apikey"], str), f"API key {dct['apikey']} must be a string"
            assert len(dct["apikey"]) > 0, f"API key must not be empty"

        if "start" in dct:
            dct["start"] = constants._validate_date(dct["start"])
        if "end" in dct:
            dct["end"] = constants._validate_date(dct["end"])
        if "forecast_date" in dct:
            dct["forecast_date"] = constants._validate_date(dct["forecast_date"])

        if "version" in dct:
            assert (
                dct["version"] in constants.MODEL_VERSIONS
            ), f"Version {dct['version']} not in {constants.MODEL_VERSIONS}"

        if "shapefile" in dct and "debias" in dct and dct["debias"]:
            raise ValueError("Cannot debias with shapefile locations")

        return dct

    def asargs(self, **kwargs) -> str:
        """Render as api arguments.

        Encodes a dictionary of the type returned by `asdict` into a
        single string appropriate for an API request.
        """
        return urllib.parse.urlencode(self.asdict(**kwargs), safe=",")

    def __validate(self):
        if self.location_file:
            assert os.path.exists(
                self.location_file
            ), f"Location file {self.location_file} does not exist"
            assert not self.lat, "Cannot specify both lat and location_file"
            assert not self.lon, "Cannot specify both lon and location_file"
            assert not self.region, "Cannot specify both region and location_file"
            assert not self.shapefile, "Cannot specify both shape_file and location_file"
        elif self.shapefile:
            assert os.path.exists(self.shapefile), f"Shape file {self.shapefile} does not exist"
            assert not self.region, "Cannot specify both region and shapefile"
            assert not self.lat, "Cannot specify both lat and shape_file"
            assert not self.lon, "Cannot specify both lon and shape_file"
        elif self.region:
            assert not self.lat, "Cannot specify both lat and region"
            assert not self.lon, "Cannot specify both lon and region"
            assert not self.location_file, "Cannot specify both location_file and region"
            assert not self.shapefile, "Cannot specify both shape_file and region"
        else:
            assert self.lat, "Must specify lat & lon, location_file, shapefile, or region"
            assert self.lon, "Must specify lat & lon, location_file, shapefile, or region"
            assert -90 <= self.lat <= 90, "Latitude must be between -90 and 90 degrees"
            assert -180 <= self.lon <= 180, "Longitude must be between -180 and 180 degrees"

    def __str__(self):
        """Return a string representation of the Location object."""
        if self.location_file:
            return f"location file: {self.location_file}"
        elif self.shapefile:
            return f"shape file: {self.shapefile}"
        elif self.region:
            return f"region: {self.region}"
        else:
            return f"({self.lat}, {self.lon})"

    def __eq__(self, other):
        """Check if two Location objects are equal."""
        if self.location_file:
            return self.location_file == other.location_file
        elif self.shapefile:
            return self.shapefile == other.shape_file
        elif self.region:
            return self.region == other.region
        else:
            return self.lat == other.lat and self.lon == other.lon

    def __ne__(self, other):
        """Check if two Location objects are not equal."""
        return not self.__eq__(other)



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Interface to the Salient `upload_file` API.

Command line usage example:
```
cd ~/salientsdk
python -m salientsdk.upload_file
```
"""

import json
import os

import pandas as pd
import requests

from .constants import _build_url
from .login_api import get_current_session, login


def upload_file(
    file: str, verbose: bool = True, session: requests.Session = get_current_session()
) -> None:
    """Uploads a geography file to the Salient API.

    An interface to to the Salient
    [upload_file](https://api.salientpredictions.com/v2/documentation/api/#/General/upload_file)
    API endpoint.

    Args:
        file (str): the file to upload (e.g. a shapefile or CSV).
        verbose (bool): whether to print status messages.
        session (requests.Session): the session to use for the upload.

    """
    if verbose:
        print(f"Uploading {file}")

    (url, loc_file) = _build_url("upload_file")

    # do we need to open .zipped shapefiles in 'rb' binary mode?
    req = session.post(url, files={"file": open(file, "r")})
    req.raise_for_status()
    if verbose:
        print(req.text)

    return None


def upload_bounding_box(
    north: float,
    south: float,
    east: float,
    west: float,
    geoname: str,
    force: bool = False,
    verbose: bool = True,
    session: requests.Session = get_current_session(),
) -> str:
    """Upload a bounding box.

    Create and upload a GeoJSON shapefile with a rectangular bounding box
    for later use with the `shapefile` location argument.

    Args:
        north (float): Northern extent decimal latitude
        south (float): Southern extent decimal latitude
        east (float): Eastern extent decimal longitude
        west (float): Western extent decimal longitude
        geoname (str): Name of the GeoJSON file and object to create
        force (bool): If the file already exists, don't upload it
        verbose (bool): Whether to print status messages
        session (requests.Session): The session object to use for the request

    Returns:
        str: File name of the GeoJSON file
    """
    geofile = geoname + ".geojson"
    if not force and os.path.exists(geofile):
        if verbose:
            print(f"File {geofile} already exists")
        return geofile

    assert west < east, "West must be less than East"
    assert south < north, "South must be less than North"
    assert session is not None, "login to create a session"

    geoshape = {
        "type": "Feature",
        "properties": {"name": geoname},
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [[west, north], [east, north], [east, south], [west, south], [west, north]]
            ],
        },
    }
    with open(geofile, "w") as f:
        json.dump(geoshape, f)

    upload_file(geofile, verbose, session)

    return geofile


def upload_location_file(
    lats: list[float],
    lons: list[float],
    names: list[str],
    geoname: str,
    force: bool = False,
    verbose: bool = False,
    session: requests.Session = get_current_session(),
    **kwargs,
) -> str:
    """Upload a vector of locations.

    Create and upload a CSV file with a list of locations for
    later use with the `location_file` location argument.

    Args:
        lats (list[float]): List of decimal latitudes
        lons (list[float]): List of decimal longitudes
        names (list[str]): List of names for the locations
        geoname (str): Name of the CSV file and object to create
        force (bool): When False, if the file already exists don't upload it
        verbose (bool): If True, print status messages
        session (requests.Session): The session object to use for the request
        **kwargs: Additional columns to include in the CSV file

    Returns:
        str: File name of the CSV file
    """
    geofile = geoname + ".csv"
    if not force and os.path.exists(geofile):
        if verbose:
            print(f"File {geofile} already exists")
        return geofile

    loc_table = pd.DataFrame({"lat": lats, "lon": lons, "name": names, **kwargs})
    loc_table.to_csv(geofile, index=False)

    upload_file(geofile, verbose, session)

    return geofile


if __name__ == "__main__":
    session = login()

    geofile = upload_bounding_box(
        west=-109.1, east=-102.0, south=37.8, north=41.0, geoname="Colorado", session=session
    )

    # read the JSON geofile
    with open(geofile, "r") as f:
        print(json.load(f))



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Historical data timeseries.

This module is an interface to the Salient `data_timeseries` API, which returns historical
observed data.  It also includes utility functions for operating on the returned data.

Command line usage example:

```
cd ~/salientsdk
# this will get a single variable in a single file:
python -m salientsdk data_timeseries -lat 42 -lon -73 -fld all --start 2020-01-01 --end 2020-12-31 --force
# this will get multiple variables in separate files:
python -m salientsdk data_timeseries -lat 42 -lon -73 -fld all -var temp,precip
```

"""

import os

import requests
import xarray as xr

from .constants import _build_url
from .location import Location
from .login_api import download_query, get_api_key, get_current_session, get_verify_ssl


def data_timeseries(
    loc: Location,
    variable: str = "temp",
    field: str = "anom",
    debias: bool = False,
    start: str = "1950-01-01",
    end: str = "-today",
    format: str = "nc",
    frequency: str = "daily",
    force: bool = False,
    session: requests.Session = get_current_session(),
    apikey: str | None = get_api_key(),
    verify: bool = get_verify_ssl(),
    verbose: bool = False,
    **kwargs,
) -> str | dict[str, str]:
    """Get a historical time series of ERA5 data.

    This function is a convenience wrapper to the Salient
    [API](https://api.salientpredictions.com/v2/documentation/api/#/Historical/get_data_timeseries).

    Args:
        loc (Location): The location to query
        variable (str): The variable to query, defaults to `temp`
            To request multiple variables, separate them with a comma `temp,precip`
            This will download one file per variable
            See the
            [Data Fields](https://salientpredictions.notion.site/Variables-d88463032846402e80c9c0972412fe60)
            documentation for a full list of available historical variables.
        field (str): The field to query, defaults to "anom"
        debias (bool): If True, debias the data to local observations.  Disabled for `shapefile` locations.  [detail](https://salientpredictions.notion.site/Debiasing-2888d5759eef4fe89a5ba3e40cd72c8f)
        start (str): The start date of the time series
        end (str): The end date of the time series
        format (str): The format of the response
        frequency (str): The frequency of the time series
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided
            and `get_api_key()` returns `None`.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        **kwargs: Additional arguments to pass to the API

    Keyword Arguments:
        units (str): `SI` or `US`

    Returns:
        str | dict: the file name of the downloaded data.  File names are a hash of the query parameters.
            When `force=False` and the file already exists, the function will return the file name
            almost instantaneously without querying the API.
            If multiple variables are requested, returns a `dict` of `{variable:file_name}`
    """
    assert field in [
        "anom",
        "anom_d",
        "anom_ds",
        "anom_qnt",
        "anom_s",
        "clim",
        "stdv",
        "trend",
        "vals",
        "all",
    ], f"Invalid field {field}"
    assert format in ["nc", "csv"], f"Invalid format {format}"
    assert frequency in [
        "daily",
        "weekly",
        "monthly",
        "3-monthly",
    ], f"Invalid frequency {frequency}"

    # if there is a comma in variable, vectorize:
    if isinstance(variable, str) and "," in variable:
        variable = variable.split(",")

    if isinstance(variable, list):
        file_names = {
            var: data_timeseries(
                loc=loc,
                variable=var,
                field=field,
                debias=debias,
                start=start,
                end=end,
                format=format,
                frequency=frequency,
                force=force,
                session=session,
                verify=verify,
                verbose=verbose,
                apikey=apikey,
                **kwargs,
            )
            for var in variable
        }
        if verbose:
            print(file_names)
        return file_names

    endpoint = "data_timeseries"
    args = loc.asdict(
        start=start,
        end=end,
        debias=debias,
        field=field,
        format=format,
        frequency=frequency,
        variable=variable,
        apikey=apikey,
        **kwargs,
    )

    (query, file_name) = _build_url(endpoint, args)

    download_query(
        query=query,
        file_name=file_name,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
    )

    return file_name


def load_multihistory(files: dict, fields: list[str] = ["vals"]) -> xr.Dataset:
    """Load multiple history files and merge them into a single dataset.

    Args:
        files (dict): Dictionary of `{variable:file_name}` of the type returned by
                      `data_timeseries` when multiple `variable`s are requested
                      e.g. `data_timeseries(..., variable = "temp,precip")`
        fields (list[str]): List of fields to extract from the history files

    Returns:
        xr.Dataset: The merged dataset, where each field and variable is renamed
        to `<variable>_<field>` or simply `variable` if field = "vals".
    """

    def __extract_history_fields(file: str, variable: str, fields: str) -> xr.Dataset:
        hst = xr.load_dataset(file)
        hst = hst[fields]
        fields_new = [variable if field == "vals" else variable + "_" + field for field in fields]
        hst = hst.rename({field: field_new for field, field_new in zip(fields, fields_new)})
        for fld in fields_new:
            hst[fld].attrs = hst.attrs
        hst.attrs = {}
        hst.close()

        return hst

    # Would prefer to use xr.open_mfdataset, but we need to pass in the variable name
    # Can convert when history files have a short_name attribute
    # https://salientpredictions.atlassian.net/browse/RD-1184
    hst = xr.merge(
        [__extract_history_fields(files[variable], variable, fields) for variable in files.keys()]
    )
    return hst



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Interface to the Salient data_timeseries API.

Command line usage example:
```
cd ~/salientsdk
python -m salientsdk downscale -lat 42 -lon -73 --date 2020-01-01 -u username -p password --force
```

"""

import os
from datetime import datetime

import requests
import xarray as xr

from .constants import _build_url, get_model_version
from .location import Location
from .login_api import download_query, get_api_key, get_current_session, get_verify_ssl

REFERENCE_CLIMS = ["30_yr", "10_yr", "5_yr"]


def downscale(
    loc: Location,
    variables: str = "temp,precip",
    debias: bool = False,
    date: str = "-today",
    reference_clim: str = "30_yr",
    members: int = 50,
    version: str = get_model_version(),
    force: bool = False,
    session: requests.Session = get_current_session(),
    apikey: str | None = get_api_key(),
    verify: bool = get_verify_ssl(),
    verbose: bool = False,
    **kwargs,
) -> str | list[str]:
    """Temporal downscale of forecasts.

    Convert temporally coarse probabilistic forecasts into granular daily ensembles.
    For more detail, see the
    [api doc](https://api.salientpredictions.com/v2/documentation/api/#/Forecasts/downscale).

    Args:
        loc (Location): The location to query
        variables (str): The variables to query, separated by commas
            See the
            [Data Fields](https://salientpredictions.notion.site/Variables-d88463032846402e80c9c0972412fe60)
            documentation for a full list of available variables.
        date (str): The start date of the time series.
        reference_clim (str): Reference period to calculate anomalies.
        members (int): The number of ensemble members to download
        debias (bool): If True, debias the data
        version (str): The model version of the Salient `blend` forecast.
        force (bool): If False (default), don't download the data if it already exists
        session (requests.Session): The session object to use for the request
        apikey (str | None): The API key to use for the request.
            In most cases, this is not needed if a `session` is provided
            and `get_api_key()` returns `None`.
        verify (bool): If True (default), verify the SSL certificate
        verbose (bool): If True (default False) print status messages
        **kwargs: Additional arguments to pass to the API

    Keyword Arguments:
        gdd_base (int): The base temperature for growing degree days
        units (str): US or SI

    Returns:
        str: the name of the downloaded (or cached) downscale data file
    """
    format = "nc"
    frequency = "daily"
    model = "blend"

    assert members > 0, "members must be a positive integer"
    assert isinstance(session, requests.Session), "Must login to the Salient API first"

    if date == "-today":
        date = datetime.today().strftime("%Y-%m-%d")

    assert reference_clim in REFERENCE_CLIMS, f"reference_clim must be one of {REFERENCE_CLIMS}"

    endpoint = "downscale"
    args = loc.asdict(
        variables=variables,
        date=date,
        reference_clim=reference_clim,
        debias=debias,
        members=members,
        format=format,
        model=model,
        version=version,
        frequency=frequency,
        apikey=apikey,
        **kwargs,
    )

    (query, file_name) = _build_url(endpoint, args)

    download_query(
        query=query,
        file_name=file_name,
        force=force,
        session=session,
        verify=verify,
        verbose=verbose,
        format=format,
    )

    return file_name
